import contextlib
import os
import toml

from pkg_resources import resource_filename
from typing import Optional
from .env_utils import EnvUtils
from .misc_utils import classproperty, local_attrs


class ProjectRegistry:

    SHOW_HERALD_WHEN_INITIALIZED = True

    REGISTERED_PROJECTS = {}

    # All of these might never be other than None so be careful when accessing them.
    APPLICATION_PROJECT_HOME = None
    PYPROJECT_TOML_FILE = None
    PYPROJECT_TOML = None
    POETRY_DATA = None
    # This is expected to ultimately be set properly.
    _PYPROJECT_NAME = None

    @classproperty
    def PYPROJECT_NAME(cls) -> str:  # noQA - PyCharm thinks this should be 'self'
        if cls._PYPROJECT_NAME is None:
            cls.initialize_pyproject_name()
        result: Optional[str] = cls._PYPROJECT_NAME
        if result is None:
            raise ValueError(f"ProjectRegistry.PROJECT_NAME not initialized properly.")
        return result

    @classmethod
    def initialize_pyproject_name(cls, project_home=None, pyproject_toml_file=None):
        if cls._PYPROJECT_NAME is None:
            # This isn't the home of snovault, but the home of the snovault-based application.
            # So in CGAP, for example, this would want to be the home of the CGAP application.
            # If not set, it will be assumed that the current working directory is that.
            if not project_home:
                project_home = os.environ.get("APPLICATION_PROJECT_HOME", os.path.abspath(os.curdir))
            cls.APPLICATION_PROJECT_HOME = project_home
            if not pyproject_toml_file:
                expected_pyproject_toml_file = os.path.join(project_home, "pyproject.toml")
                pyproject_toml_file = (expected_pyproject_toml_file
                                       if os.path.exists(expected_pyproject_toml_file)
                                       else None)
            cls.PYPROJECT_TOML_FILE = pyproject_toml_file
            cls.PYPROJECT_TOML = pyproject_toml = (toml.load(cls.PYPROJECT_TOML_FILE)
                                                   if cls.PYPROJECT_TOML_FILE
                                                   else None)
            cls.POETRY_DATA = (pyproject_toml['tool']['poetry']
                               if pyproject_toml
                               else None)

            declared_pyproject_name = os.environ.get("APPLICATION_PYPROJECT_NAME")
            inferred_pyproject_name = cls.POETRY_DATA['name'] if cls.POETRY_DATA else None
            if (declared_pyproject_name and inferred_pyproject_name
                    and declared_pyproject_name != inferred_pyproject_name):
                raise RuntimeError(f"APPLICATION_PYPROJECT_NAME={declared_pyproject_name!r},"
                                   f" but {pyproject_toml_file} says it should be {inferred_pyproject_name!r}")

            cls._PYPROJECT_NAME = declared_pyproject_name or inferred_pyproject_name

    @classmethod
    def register(cls, pyproject_name):
        """
        Registers a class to be used based on pyproject_name (the name in the top of pyproject.toml).
        Note that this means that cgap-portal and fourfront will both register as 'encoded',
        as in:

            @Project.register('encoded')
            class FourfrontProject(EncodedCoreProject):
                PRETTY_NAME = "Fourfront"

        Since fourfront and cgap-portal don't occupy the same space, no confusion should result.
        """
        def _wrap_class(the_class):
            the_class_name = the_class.__name__
            if not issubclass(the_class, Project):
                raise ValueError(f"The class {the_class_name} must inherit from Project.")
            explicit_pyproject_name = the_class.__dict__.get('PYPROJECT_NAME')
            if explicit_pyproject_name:
                message = f"Explicit {the_class_name}.PYPROJECT_NAME={explicit_pyproject_name!r} is not permitted."
                if explicit_pyproject_name != pyproject_name:
                    message += f" An assignment to {pyproject_name!r} is intended to be managed implicitly."
                else:
                    message += " This assignment is intended to be managed implicitly."
                raise ValueError(message)
            else:
                the_class._PYPROJECT_NAME = pyproject_name  # no need for the caller to say this redundantly

            lower_registry_name = pyproject_name.lower()
            for x in ['cgap-portal', 'fourfront', 'smaht']:
                if x in lower_registry_name:
                    # It's an easy error to make, but the name of the project from which we're gaining foothold
                    # in pyproject.toml is 'encoded', not 'cgap-portal', etc., so the name 'encoded' will be
                    # needed for bootstrapping. So it should look like
                    # -kmp 15-May-2023
                    raise ValueError(f"Please use ProjectRegistry.register('encoded'),"
                                     f" not ProjectRegistry.register({pyproject_name!r})."
                                     f" This registration is just for bootstrapping."
                                     f" The class can still be {the_class_name}.")
            cls.REGISTERED_PROJECTS[pyproject_name] = the_class
            return the_class
        return _wrap_class

    @classmethod
    def _lookup(cls, name):
        """
        Returns the project object with the given name.

        :param name: a string name that was used in a ProjectRegistry.register decorator

        NOTE: There is no need for this function to be called outside of this class except for testing.
              Really only one of these should be instantiated per running application, and that's
              done automatically by this class.
        """
        project_class = cls.REGISTERED_PROJECTS.get(name)
        return project_class

    @classmethod
    def _make_project(cls):
        """
        Creates and returns an instantiated project object for the current project.

        The project to use can be specified by setting the environment variable APPLICATION_PROJECT_HOME
        to a particular directory that contains the pyproject.toml file to use.
        If no such variable is set, the current working directory is used.

        NOTE: There is no need for this function to be called outside of this class except for testing.
              Really only one of these should be instantiated per running application, and that's
              done automatically by this class.
        """
        project_class = cls._lookup(cls.PYPROJECT_NAME)
        assert issubclass(project_class, Project)
        project: Project = project_class()
        return project  # instantiate and return

    _app_project = None

    @classmethod
    def initialize(cls, force=False):
        if cls._app_project and not force:
            return cls._app_project
        cls._app_project = cls._make_project()
        if cls.SHOW_HERALD_WHEN_INITIALIZED:
            cls.show_herald()
        app_project: Project = cls.app_project  # Now that it's initialized, make sure it comes from the right place
        return app_project

    @classmethod
    def show_herald(cls):
        app_project = Project.app_project_maker()

        print("=" * 80)
        print(f"APPLICATION_PROJECT_HOME == {cls.APPLICATION_PROJECT_HOME!r}")
        print(f"PYPROJECT_TOML_FILE == {cls.PYPROJECT_TOML_FILE!r}")
        print(f"PYPROJECT_NAME == {cls.PYPROJECT_NAME!r}")
        the_app_project = Project.app_project
        the_app_project_class = the_app_project.__class__
        the_app_project_class_name = the_app_project_class.__name__
        assert (Project.app_project
                == app_project()
                == the_app_project_class.app_project
                == the_app_project.app_project), (
            "Project consistency check failed."
        )
        print(f"{the_app_project_class_name}.app_project == Project.app_project == app_project() == {app_project()!r}")
        print(f"app_project().NAME == {app_project().NAME!r}")
        print(f"app_project().PRETTY_NAME == {app_project().PRETTY_NAME!r}")
        print(f"app_project().PACKAGE_NAME == {app_project().PACKAGE_NAME!r}")
        print(f"app_project().APP_NAME == {app_project().APP_NAME!r}")
        print(f"app_project().APP_PRETTY_NAME == {app_project().APP_PRETTY_NAME!r}")
        print("=" * 80)

    @classproperty
    def app_project(cls):  # noQA - PyCharm thinks we should use 'self'
        """
        Once the project is initialized, ProjectRegistry.app_project returns the application object
        that should be used to dispatch project-dependent behavior.
        """
        if cls._app_project is None:
            # You need to put a call to
            raise RuntimeError(f"Attempt to access {cls.__name__}.project before .initialize() called.")
        app_project: Project = cls._app_project
        return app_project

    @classmethod
    @contextlib.contextmanager
    def project_registry_test_context(cls):
        with local_attrs(cls, REGISTERED_PROJECTS={}):
            assert cls.REGISTERED_PROJECTS == {}
            yield


class Project:
    """
    A class that should be a superclass of all classes registered using ProjectRegistry.register

    All such classes have these names:
      .NAME - The name of the project in pyproject.toml
      .PACKAGE_NAME - The pypi name of the project, useful for pkg_resources, for example.
      .PRETTY_NAME - The pretty name of the package name
      .APP_NAME - The ame of the project application (see dcicutils.common and the orchestrated app in EnvUtils)
      .APP_PRETTY_NAME - The pretty name of the project application.

    Some sample usess of pre-defined attributes of a Project that may help motivate the choice of attribute names:

      registered  |
         name     |     NAME     |  PACKAGE_NAME  |  PRETTY NAME |  APP_NAME | APP_PRETTY_NAME
     -------------+--------------+----------------+--------------+-----------+----------------
     snovault     | dcicsnovault |  snovault      | Snovault     | snovault  | Snovault
     encoded-core | encoded-core |  encoded-core  | Encoded Core | core      | Core
     encoded      | cgap-portal  |  cgap-portal   | CGAP Portal  | cgap      | CGAP
     encoded      | fourfront    |  fourfront     | Fourfront    | fourfront | Fourfront
     encoded      | smaht-portal |  smaht-portal  | SMaHT Portal | smaht     | SMaHT

     The registered name is the one used with the ProjectRegistry.register() decorator.
    """

    NAME = 'project'

    _PYPROJECT_NAME = None

    @classmethod
    def _prettify(cls, name):
        return name.title().replace("Cgap", "CGAP").replace("Smaht", "SMaHT").replace("-", " ")

    @classproperty
    def PACKAGE_NAME(cls):  # noQA - PyCharm wants the variable name to be self
        return cls.NAME.replace('dcic', '')

    @classproperty
    def PYPROJECT_NAME(cls) -> str:  # noQA - PyCharm wants the variable name to be self
        pyproject_name = cls._PYPROJECT_NAME
        if not pyproject_name:
            raise RuntimeError(f"Class {cls} was not defined with ProjectRegistry.register")
        return pyproject_name

    @classproperty
    def PRETTY_NAME(cls):  # noQA - PyCharm wants the variable name to be self
        return cls._prettify(cls.PACKAGE_NAME)

    @classproperty
    def APP_NAME(cls):  # noQA - PyCharm wants the variable name to be self
        return cls.PACKAGE_NAME.replace('-portal', '').replace('encoded-', '')

    @classproperty
    def APP_PRETTY_NAME(cls):  # noQA - PyCharm wants the variable name to be self
        return cls._prettify(cls.APP_NAME)

    @classproperty
    def app_project(cls):  # noQA - PyCharm wants the variable name to be self
        """
        Project.app_project returns the actual instantiated project for app-specific behavior,
        which might be of this class or one of its subclasses.

        This access will fail if the project has not been initialized.
        """
        return ProjectRegistry.app_project

    @classmethod
    def initialize_app_project(cls, initialize_env_utils=True):
        if initialize_env_utils:
            EnvUtils.init()
        project: Project = ProjectRegistry.initialize()
        return project

    @classmethod
    def app_project_maker(cls):

        def app_project(initialize=False, initialization_options: Optional[dict] = None):
            if initialize:
                Project.initialize_app_project(**(initialization_options or {}))
            return Project.app_project

        return app_project

    def project_filename(self, filename):
        """Returns a filename relative to given instance."""
        if self != self.app_project:
            raise RuntimeError(f"{self}.project_filename invoked,"
                               f" but {self} is not the app_project, {self.app_project}.")
        return resource_filename(self.PYPROJECT_NAME, filename)



import contextlib
import io
import json
import os

from dcicutils.exceptions import AppEnvKeyMissing, AppServerKeyMissing
from dcicutils.misc_utils import remove_suffix


_KEY_MANAGERS = {}


class KeyManager:

    APP_NAME = None
    APP_TOKEN = None  # Set this to a string to override APP_NAME.lower() as the app token

    KEYS_FILE = None
    KEYS_FILE_VAR = None

    _REGISTERED = False

    # Instance Methods

    def __init__(self, keys_file=None):
        """
        Defines what is done upon creation of a key manager instance.

        NOTE: The only such classes that are instantiable are those that have been registered.
              The reason is that the register decorator performs important initializations of class variables.

        Args:
            keys_file: the name of a filename to use instead of the name chosen by default
        """

        if not self._REGISTERED:
            raise RuntimeError("Only registered KeyManagers can be instantiated.")
        self.keys_file = keys_file or os.environ.get(self.KEYS_FILE_VAR) or self.KEYS_FILE
        if not self.keys_file:
            raise ValueError(f"No KEYS_FILE attribute in {self}, and no {self.KEYS_FILE_VAR} environment variable.")

    def get_keydicts(self) -> dict:
        """
        Parses and returns the keys file (held by self.keys_file)

        NOTE: No caching is done on a theory that you don't do this super-often, and you might have edited the
              keys file since last look, so you'd want most-up-to-date credentials. If those assumptions don't
              hold, please request additional support for this class rather than creating ad hoc solutions outside.

        Returns:
            a dictionary that maps an environment name to its auth information,
            which is a dict with keys 'key', 'secret', and 'server'.

        Raises:
            ValueError: if the file is ill-formatted (does not contain JSON or JSON is not a dictionary)
        """
        if not os.path.exists(self.keys_file):
            return {}
        with io.open(self.keys_file) as fp:
            keys = json.load(fp)
            if not isinstance(keys, dict):
                raise ValueError(f"The file {self.keys_file} did not contain a Python dictionary (in JSON format).")
            return keys

    def get_keydict_for_env(self, env):
        """
        Gets the appropriate auth info (as a dict) for talking to a given beanstalk environment.

        Args:
            env: the name of a beanstalk environment

        Returns:
            Auth information as a dict with keys 'key', 'secret', and 'server'.

        Raises:
            ValueError: if the file is ill-formatted (does not contain JSON or JSON is not a dictionary)
        """

        self._check_env(env)

        keydicts = self.get_keydicts()
        keydict = keydicts.get(env)
        if not keydict:
            raise AppEnvKeyMissing(env=env, key_manager=self)
        return keydict

    def get_keypair_for_env(self, env):
        """
        Gets the appropriate auth info (as a pair/tuple) for talking to a given beanstalk environment.

        Args:
            env: the name of an environment

        Returns:
            Auth information as a (key, secret) tuple.
        """

        return self.keydict_to_keypair(self.get_keydict_for_env(env=env))

    def get_keydict_for_server(self, server):  # tested
        """
        Gets the appropriate auth info (as a dict) for talking to a given beanstalk environment.

        Args:
            server: the name of a server

        Returns:
            Auth information.
            The auth is a keypair, though we might change this to include a JWT token in the the future.
        """

        self._check_server(server)

        keydicts = self.get_keydicts()
        server_to_find = server.rstrip('/')
        for keydict in keydicts.values():
            if keydict['server'].rstrip('/') == server_to_find:
                return keydict
        raise AppServerKeyMissing(server=server, key_manager=self)

    def get_keypair_for_server(self, server):  # tested
        """
        Gets the appropriate auth info (as a pair/tuple) for talking to a given beanstalk environment.

        Args:
            server: the name of a server

        Returns:
            Auth information as a (key, secret) tuple.
        """

        return self.keydict_to_keypair(self.get_keydict_for_server(server=server))

    # Class Methods

    @classmethod
    @contextlib.contextmanager
    def default_keys_file_for_testing(cls, filename):
        """
        Sets the default keys file for in cls.KEYS_FILE to the indicated filename.
        Ordinarily, in a non-testing environment, one would set an environment variable to do this,
        but in the testing environment that has been set already and to a value that has nothing to
        do with testing. So this bypasses the normal environment variable setup and sets it directly
        in the class and only for the duration of an evaluation context.
        """

        old_filename = cls.KEYS_FILE
        try:
            cls.KEYS_FILE = filename or cls._default_keys_file(for_testing=True)
            yield
        finally:
            cls.KEYS_FILE = old_filename

    @classmethod
    def register(cls, *, name):
        """
        A decorator that used to register a key manager class, so that it can be later looked up by the given name.
        This also does important class variable initialization for the key manager class,
        so no key manager class can be instantiated unless it has been registered in this way.

        Args:

            name:  a name token (a str) that should match the first part of the class name, before 'KeyManager'.
                   e.g., for CGAPKeyManager, the name should be 'cgap'.

        """
        def _register_class(key_manager_class):
            assert issubclass(key_manager_class, KeyManager)
            if name in _KEY_MANAGERS:
                raise ValueError(f"A KeyManager named {name!r} has already been defined.")
            key_manager_class._init_class_variables()
            key_manager_class._REGISTERED = True
            _KEY_MANAGERS[name] = cls
            return key_manager_class
        return _register_class

    @classmethod
    def _init_class_variables(cls):
        class_name = cls.__name__
        # print(f"cls={cls!r}, name={class_name!r}")
        suffix = "KeyManager"
        if not class_name.endswith(suffix):
            raise ValueError(f"The name, {class_name!r}, of a KeyManager class to be registered"
                             f" does not end in {suffix!r}.")
        app_name = remove_suffix(suffix=suffix, text=class_name, required=True)
        if cls.APP_NAME is None:
            cls.APP_NAME = app_name
        elif cls.APP_NAME.lower() != app_name.lower():
            raise ValueError(f"A KeyManager class with APP_NAME = {cls.APP_NAME!r} expects a name"
                             f" like {cls.APP_NAME}{suffix}.")
        else:
            app_name = cls.APP_NAME
        if cls.APP_TOKEN is None:
            cls.APP_TOKEN = app_token = app_name.upper()
        else:
            app_token = cls.APP_TOKEN
            if not app_token.isupper():
                raise ValueError(f"The APP_TOKEN {app_token!r} must be all-uppercase.")
            elif app_token[:1] != app_name[0].upper():
                raise ValueError(f"The APP_TOKEN {app_token!r} must have the same first letter as {app_name!r}")
        app_prefix = app_token + "_"
        if not cls.KEYS_FILE_VAR:
            cls.KEYS_FILE_VAR = f"{app_prefix}KEYS_FILE"
        elif not cls.KEYS_FILE_VAR.startswith(app_prefix):
            raise ValueError("The {class_name}.KEYS_FILE_VAR must begin with {app_prefix!r}.")
        if not cls.KEYS_FILE:
            cls.KEYS_FILE = cls._default_keys_file()
        # print(f"Defining {class_name} "
        #       f"with APP_NAME={app_name!r} APP_TOKEN={app_token!r} KEYS_FILE_VAR={cls.KEYS_FILE_VAR!r}.")

    @classmethod
    def create(cls, name, **kwargs):
        """
        Allows creation of a registered key manager class using an alternate protocol that uses name lookup.
        For example, the following two are equivalent ways of creating an instance:
            CGAPKeyManager(**kwargs)
            KeyManager.create('cgap', **kwargs)
        """

        key_manager_class = _KEY_MANAGERS.get(name)
        if not key_manager_class:
            raise ValueError(f"There is no registered KeyManager class named {name!r}.")
        return key_manager_class(**kwargs)

    @classmethod
    def _default_keys_file(cls, for_testing=False):
        suffix = "-for-testing" if for_testing else ""
        return os.path.expanduser(f"~/.{cls.APP_TOKEN.lower()}-keys{suffix}.json")

    @classmethod
    def keypair_to_keydict(cls, auth_tuple, *, server):
        """
        Translates an auth tuple (key, secret) to a keydict with {"key": ..., "secret": ..., "server": ...}.
        Since the tuple must contain a secret and the keydict requires it, a server argument must be provided.
        That argument is required to be passed as a keyword.

        >>> KeyManager.keypair_to_keydict(('my-key', 'abra-cadabra'), server='http://whatever')
        {"key": "my-key", "secret": "abra-cadabra", "server": "http://whatever"}

        Args:
            auth_tuple (tuple): an auth tuple of the form (key, secret)
            server (str): a server name, such as "https://cgap.hms.harvard.edu" or "http://localhost:8000"
        Returns:
            a keydict of the form {"key": ..., "secret": ..., "server": ...}
        """

        auth_dict = {
            'key': auth_tuple[0],
            'secret': auth_tuple[1],
            'server': server,
        }
        return auth_dict

    @classmethod
    def keydict_to_keypair(cls, auth_dict):
        """
        Translates a keydict with {"key": ..., "secret": ..., "server": ...} to an auth tuple (key, secret).
        Since the tuple does not have a place for the server, it will be discarded.

        >>> KeyManager.keydict_to_keypair({"key": "my-key", "secret": "abra-cadabra", "server": "http://whatever"})
        ('my-key', 'abra-cadabra')

        Args:
            auth_dict (dict): a dictionary with entries for 'key' and 'secret'; any 'server' will be ignored
        Returns:
            a keypair of the form (key, secret)

        """
        return (
            auth_dict['key'],
            auth_dict['secret']
        )

    @classmethod
    def _check_env(cls, env):
        if not env or not isinstance(env, str):
            raise ValueError(f"Not a valid environment: {env}")

    @classmethod
    def _check_server(cls, server):
        if not server or not isinstance(server, str):
            raise ValueError(f"Not a valid server: {server}")


@KeyManager.register(name='cgap')
class CGAPKeyManager(KeyManager):
    pass


@KeyManager.register(name='fourfront')
class FourfrontKeyManager(KeyManager):
    pass



# Script to publish the Python package in the CURRENT git repo to PyPi.
# Does the following checks before allowing a publish:
#
# 1. The git repo MUST NOT contain unstaged changes.
# 2. The git repo MUST NOT contain staged but uncommitted changes.
# 3. The git repo MUST NOT contain committed but unpushed changes.
# 4. The git repo package directories MUST NOT contain untracked files,
#    OR if they do contain untracked files then you must confirm this is OK.
# 5. The version being published must NOT have already been published.
#
# ASSUMES you have these environment variables correctly set for PyPi publishing:
#
# - PYPI_USER
# - PYPI_PASSWORD
#
# Prompts for yes or no before publish is actually done. There is a --noconfirm
# option to skip this confimation, however it is only allowed when running in the
# context of GitHub actions - it checks for the GITHUB_ACTIONS environment variable.
#
# FYI: This was created late April 2023 after a junk file containing development
# logging output containing passwords was accidentally published to PyPi;
# item #4 above specifically addresses/prevents this. Perhaps better
# would be if publishing only happened via GitHub actions.

import argparse
import os
import requests
import subprocess
import toml
from typing import Tuple, Union


DEBUG = False
PRINT = print


def main() -> None:

    def is_github_actions_context():
        return "GITHUB_ACTIONS" in os.environ

    argp = argparse.ArgumentParser()
    argp.add_argument("--noconfirm", required=False, dest="noconfirm", action="store_true")
    argp.add_argument("--debug", required=False, dest="debug", action="store_true")
    args = argp.parse_args()

    if args.debug:
        global DEBUG
        DEBUG = True

    if args.noconfirm and not is_github_actions_context():
        PRINT("The --noconfirm flag is only allowed within GitHub actions!")
        exit_with_no_action()

    if not verify_git_repo():
        exit_with_no_action()

    if not verify_unstaged_changes():
        exit_with_no_action()

    if not verify_uncommitted_changes():
        exit_with_no_action()

    if not verify_unpushed_changes():
        exit_with_no_action()

    if not verify_tagged():
        exit_with_no_action()

    if not verify_untracked_files():
        exit_with_no_action()

    package_name = get_package_name()
    package_version = get_package_version()

    if not verify_not_already_published(package_name, package_version):
        exit_with_no_action()

    if not args.noconfirm:
        if not answered_yes_to_confirmation(f"Do you want to publish {package_name} {package_version} to PyPi?"):
            exit_with_no_action()

    PRINT(f"Publishing {package_name} {package_version} to PyPi ...")

    if not publish_package():
        exit_with_no_action()

    PRINT(f"Publishing {package_name} {package_version} to PyPi complete.")


def publish_package(pypi_username: str = None, pypi_password: str = None) -> bool:
    if not pypi_username:
        pypi_username = os.environ.get("PYPI_USER")
    if not pypi_password:
        pypi_password = os.environ.get("PYPI_PASSWORD")
    if not pypi_username or not pypi_password:
        PRINT(f"No PyPi credentials. You must have PYPI_USER and PYPI_PASSWORD environment variables set.")
        return False
    poetry_publish_command = [
        "poetry", "publish",
        "--no-interaction", "--build",
        f"--username={pypi_username}", f"--password={pypi_password}"
    ]
    poetry_publish_results, status_code = execute_command(poetry_publish_command)
    PRINT("\n".join(poetry_publish_results))
    if status_code != 0:
        PRINT(f"Publish to PyPi failed!")
        return False
    return True


def verify_git_repo() -> bool:
    """
    If this (the current directory) looks like a git repo then return True,
    otherwise prints a warning and returns False.
    """
    _, status = execute_command("git rev-parse --is-inside-work-tree")
    if status != 0:
        PRINT("You are not in a git repo directory!")
        return False
    return True


def verify_unstaged_changes() -> bool:
    """
    If the current git repo has no unstaged changes then returns True,
    otherwise prints a warning and returns False.
    """
    git_diff_results, _ = execute_command(["git", "diff"])
    if git_diff_results:
        PRINT("You have changes to this branch that you have not staged for commit.")
        return False
    return True


def verify_uncommitted_changes() -> bool:
    """
    If the current git repo has no staged but uncommitted changes then returns True,
    otherwise prints a warning and returns False.
    """
    git_diff_staged_results, _ = execute_command(["git", "diff", "--staged"])
    if git_diff_staged_results:
        PRINT("You have changes to this branch that you have staged but not committed.")
        return False
    return True


def verify_unpushed_changes() -> bool:
    """
    If the current git repo committed but unpushed changes then returns True,
    otherwise prints a warning and returns False.
    """
    git_uno_results, _ = execute_command(["git", "status", "-uno"], lines_containing="is ahead of")
    if git_uno_results:
        PRINT("You have committed changes to this branch that you committed but not pushed.")
        return False
    return True


def verify_tagged() -> bool:
    """
    If the current git repo has a tag as its most recent commit then returns True,
    otherwise prints a warning and returns False.
    """
    git_most_recent_commit, _ = execute_command(["git", "log", "-1", "--decorate"], lines_containing="tag:")
    if not git_most_recent_commit:
        PRINT("You can only publish a tagged commit.")
        return False
    return True


def verify_untracked_files() -> bool:
    """
    If the current git repo has no untracked files then returns True,
    otherwise prints a warning, and with the list of untracked files,
    and prompts the user for a yes/no confirmation on whether or to
    continue, and returns True for a yes response, otherwise returns False.
    """
    untracked_files = get_untracked_files()
    if untracked_files:
        PRINT(f"WARNING: You are about to PUBLISH the following ({len(untracked_files)})"
              f" UNTRACKED file{'' if len(untracked_files) == 1 else 's' } -> SECURITY risk:")
        for untracked_file in untracked_files:
            PRINT(f"-- {untracked_file}")
        PRINT("DO NOT continue UNLESS you KNOW what you are doing!")
        if not answered_yes_to_confirmation("Do you really want to continue?"):
            return False
    return True


def verify_not_already_published(package_name: str, package_version: str) -> bool:
    """
    If the given package and version has not already been published to PyPi then returns True,
    otherwise prints a warning and returns False.
    """
    response = requests.get(f"https://pypi.org/project/{package_name}/{package_version}/")
    if response.status_code == 200:
        PRINT(f"Package {package_name} {package_version} has already been published to PyPi.")
        return False
    return True


def get_untracked_files() -> list:
    """
    Returns a list of untracked files for the current git repo; empty list of no untracked changes.
    """
    package_directories = get_package_directories()
    untracked_files = []
    for package_directory in package_directories:
        git_status_results, _ = execute_command(["git", "status", "-s", package_directory])
        for git_status_result in git_status_results:
            if git_status_result and git_status_result.startswith("??"):
                untracked_file = git_status_result[2:].strip()
                if untracked_file:
                    untracked_files.append(untracked_file)
    return untracked_files


def get_package_version() -> str:
    """
    Returns the tag name of the most recently created tag in the current git repo.
    """
    tag_commit, _ = execute_command("git rev-list --tags --max-count=1")
    tag_name, _ = execute_command(f"git  describe --tags  {tag_commit[0]}")
    package_version = tag_name[0]
    if package_version.startswith("v"):
        package_version = package_version[1:]
    return package_version


def get_package_name() -> str:
    """
    Returns the base name of the current git repo name.
    """
    package_name, _ = execute_command("git config --get remote.origin.url".split(" "))
    package_name = os.path.basename(package_name[0])
    if package_name.endswith(".git"):
        package_name = package_name[:-4]
    return package_name


def get_package_directories() -> list:
    """
    Returns a list of directories constituting the Python package of the current repo,
    according to the pyproject.toml file.
    """
    package_directories = []
    with open("pyproject.toml", "r") as f:
        pyproject_toml = toml.load(f)
    pyproject_package_directories = pyproject_toml["tool"]["poetry"]["packages"]
    for pyproject_package_directory in pyproject_package_directories:
        package_directory = pyproject_package_directory.get("include")
        if package_directory:
            package_directory_from = pyproject_package_directory.get("from")
            if package_directory_from:
                package_directory = os.path.join(package_directory_from, package_directory)
            package_directories.append(package_directory)
    return package_directories


def execute_command(command_argv: Union[list, str], lines_containing: str = None) -> Tuple[list, int]:
    """
    Executes the given command as a command-line subprocess, and returns a tuple whose first element
    is the list of lines from the output of the command, and the second element is the exit status code.
    """
    def cleanup_funny_output(output: str) -> str:
        return output.replace("('", "").replace("',)", "").replace("\\n\\n", "\n").replace("\\n", "\n")

    if isinstance(command_argv, str):
        command_argv = [arg for arg in command_argv.split(" ") if arg.strip()]
    if DEBUG:
        PRINT(f"DEBUG: {' '.join(command_argv)}")
    result = subprocess.run(command_argv, stdout=subprocess.PIPE,
                            stderr=subprocess.STDOUT)
    lines = result.stdout.decode("utf-8").split("\n")
    if lines_containing:
        lines = [line for line in lines if lines_containing in line]
    return [cleanup_funny_output(line.strip()) for line in lines if line.strip()], result.returncode


def answered_yes_to_confirmation(message: str) -> bool:
    """
    Prompts the user with the given message and asks for a yes or no answer,
    and if yes is the user response then returns True, otherwise returns False.
    """
    response = input(f"{message} [yes | no]: ").lower()
    if response == "yes":
        return True
    return False


def exit_with_no_action() -> None:
    """
    Exits this process immediately with status 1;
    first prints a message saying no action was taken.
    """
    PRINT("Exiting without taking action.")
    exit(1)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': '.', 'scripts': 'dcicutils/scripts'}

packages = \
['dcicutils', 'dcicutils.scripts', 'scripts']

package_data = \
{'': ['*'], 'dcicutils': ['kibana/*']}

install_requires = \
['PyJWT>=2.6.0,<3.0.0',
 'PyYAML>=5.1,<5.5',
 'aws-requests-auth>=0.4.2,<1',
 'boto3>=1.17.39,<2.0.0',
 'botocore>=1.20.39,<2.0.0',
 'docker>=4.4.4,<5.0.0',
 'elasticsearch==7.13.4',
 'gitpython>=3.1.2,<4.0.0',
 'opensearch-py>=2.0.1,<3.0.0',
 'pyOpenSSL>=23.1.1,<24.0.0',
 'pytz>=2020.4',
 'redis>=4.5.1,<5.0.0',
 'requests>=2.21.0,<3.0.0',
 'rfc3986>=1.4.0,<2.0.0',
 'structlog>=19.2.0,<20.0.0',
 'toml>=0.10.1,<1',
 'tqdm>=4.65.0,<5.0.0',
 'typing-extensions>=3.8',
 'urllib3>=1.26.6,<2.0.0',
 'webtest>=2.0.34,<3.0.0']

entry_points = \
{'console_scripts': ['publish-to-pypi = '
                     'dcicutils.scripts.publish_to_pypi:main']}

setup_kwargs = {
    'name': 'dcicutils',
    'version': '7.4.2.1b4',
    'description': 'Utility package for interacting with the 4DN Data Portal and other 4DN resources',
    'long_description': '=====\nutils\n=====\n\nCheck out our full documentation `here <https://dcic-utils.readthedocs.io/en/latest/>`_\n\nThis repository contains various utility modules shared amongst several projects in the 4DN-DCIC. It is meant to be used internally by the DCIC team and externally as a Python API to `Fourfront <https://data.4dnucleome.org>`_\\ , the 4DN data portal.\n\npip installable as the ``dcicutils`` package with: ``pip install dcicutils``\n\nSee `this document <https://dcic-utils.readthedocs.io/en/latest/getting_started.html>`_ for tips on getting started. `Go here <https://dcic-utils.readthedocs.io/en/latest/examples.html>`_ for examples of some of the most useful functions.\n\n\n.. image:: https://travis-ci.org/4dn-dcic/utils.svg?branch=master\n   :target: https://travis-ci.org/4dn-dcic/utils\n   :alt: Build Status\n\n\n.. image:: https://coveralls.io/repos/github/4dn-dcic/utils/badge.svg?branch=master\n   :target: https://coveralls.io/github/4dn-dcic/utils?branch=master\n   :alt: Coverage\n\n.. image:: https://readthedocs.org/projects/dcic-utils/badge/?version=latest\n   :target: https://dcic-utils.readthedocs.io/en/latest/?badge=latest\n   :alt: Documentation Status\n',
    'author': '4DN-DCIC Team',
    'author_email': 'support@4dnucleome.org',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/4dn-dcic/utils',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<3.10',
}


setup(**setup_kwargs)
